/***********************************************************************

	File:	bsd.c
	Rev:	a-2
	Date:	10/04/98

	Copyright (c) 1997, 1999 by David M. Warme

************************************************************************

	Bottleneck Steiner Distance stuff.

************************************************************************

	Modification Log:

	a-1:	09/04/97	warme
		: Split off from coarse.c.
	a-2:	10/04/98	warme
		: Completely re-coded to conserve memory and provide
		:  multiple implementations that provide different
		:  space-time tradeoffs.

************************************************************************/

#include "bsd.h"
#include "steiner.h"


/*
 * Global Routines
 */

dist_t			bsd (struct bsd *, int, int);
struct bsd *		compute_bsd (int, struct edge *, int);
void			shutdown_bsd (struct bsd *);


/*
 * Local Routines
 */

static struct mstadj **	build_adjacency_list (int, int, int, struct edge *);
static void		walk (int,
			      int,
			      struct bsd *,
			      int *);

/*
 * This routine returns the Bottleneck Steiner Distance between
 * two terminals, specified by index.
 */

	dist_t
bsd (

struct bsd *	bsdp,		/* IN - BSD data structure */
int		i,		/* IN - first terminal */
int		j		/* IN - second terminal */
)
{
int		n;

	n = bsdp -> n;
	if ((i < 0) OR (i >= n) OR (j < 0) OR (j >= n)) {
		fatal ("bsd: Bug 1.");
	}

	if (i EQ j) return (0.0);

	switch (bsdp -> method) {
	case 1:	/* Complete lower-triangular matrix. */
		if (i > j) {
			i = ((i * (i - 1)) >> 1) + j;
		}
		else {
			i = ((j * (j - 1)) >> 1) + i;
		}
		j = bsdp -> ematrix [i];
		return (bsdp -> mst_edges [j].len);

	case 2:	/* Cache of BSD rows. */
		/* Not yet implemented! */
		fatal ("bsd: Bug 2.");
		break;

	default:
		fatal ("bsd: Bug 1.");
		break;
	}

	return (0.0);
}

/*
 * This routine initializes the BSD data structures according to the
 * given point set and implementation method.  The first thing is to
 * compute the actual minimum spanning tree.  The rest of the initialization
 * is method-specific.
 */

	struct bsd *
compute_bsd (

int			nedges,		/* IN - number of MST edges */
struct edge *		mst_edges,	/* IN - edges of the MST */
int			method		/* IN - implementation method */
)
{
int			i;
int			j;
int			nterms;
struct bsd *		bsdp;
int16u *		rowp;
int *			tmprow;

	nterms = nedges + 1;

	/* Allocate and zero the BSD structure. */
	bsdp = NEW (struct bsd);
	memset (bsdp, 0, sizeof (*bsdp));

	if (method EQ 0) {
		/* Default is to use the full matrix for 4000 points or	*/
		/* less.  Use the cache of rows for larger problems.	*/
#if 0
		method = (nterms <= 4000) ? 1 : 2;
#else
		method = 1;
#endif
	}
	bsdp -> method		= method;

	bsdp -> n		= nterms;
	bsdp -> mst_edges	= NEWA (nedges + 1, struct edge); /* 1 extra! */

	/* Add initial zero-length edge, so that we have an edge number	*/
	/* that represents a BSD of zero.				*/
	bsdp -> mst_edges [0].len	= 0.0;
	bsdp -> mst_edges [0].p1	= 0;
	bsdp -> mst_edges [0].p2	= 0;

	memcpy (&(bsdp -> mst_edges [1]),
		mst_edges,
		nedges * sizeof (mst_edges [0]));

	bsdp -> adj_list = build_adjacency_list (nterms,
						 1,
						 nedges + 1,
						 bsdp -> mst_edges);

	bsdp -> mark = NEWA (nterms, bool);
	for (i = 0; i < nterms; i++) {
		bsdp -> mark [i] = FALSE;
	}

	switch (method) {
	case 1:	/* Complete lower triangular matrix. */

		bsdp -> ematrix = NEWA (nterms * (nterms - 1) >> 1, int16u);
		tmprow = NEWA (nterms, int);

		/* Fill in the matrix, one row at a time... */
		rowp = bsdp -> ematrix;
		for (i = 0; i < nterms; i++) {
			walk (i, 0, bsdp, tmprow);
			for (j = 0; j < i; j++) {
				*rowp++ = tmprow [j];
			}
		}
		free ((char *) tmprow);
		break;

	case 2:
		fatal ("compute_bsd: Bug 2.");
		break;

	default:
		fatal ("compute_bsd: Bug 3.");
		break;
	}

	return (bsdp);
}

/*
 * This routine converts a list of edges into a full graph structure
 * represented in adjacency list form.  The adjacency list is in two parts:
 * an array of pointers indexed by node number, and an array of "adj"
 * structures indexed by these pointers.  To find all neighbors of node K,
 * look at every "mstadj" structure between adj_list [K] and adj_list [K+1].
 */

	static
	struct mstadj **
build_adjacency_list (

int			n,		/* IN - number of nodes */
int			first_edge,	/* IN - first edge number */
int			nedges,		/* IN - number of edges */
struct edge *		edges		/* IN - array of edges */
)
{
int			i;
struct edge *		ep;
struct mstadj *		ap;
int *			count;
struct mstadj **	tmp_ptr;
struct mstadj **	adj_list;

	adj_list	= NEWA (n + 1, struct mstadj *);
	ap		= NEWA (2 * nedges, struct mstadj);
	tmp_ptr		= NEWA (n, struct mstadj *);
	count		= NEWA (n, int);

	for (i = 0; i < n; i++) {
		count [i] = 0;
	}

	/* Count neighbors for each node. */
	ep = &edges [first_edge];
	for (i = first_edge; i < nedges; ep++, i++) {
		++(count [ep -> p1]);
		++(count [ep -> p2]);
	}

	/* Set up the pointers for each node. */
	for (i = 0; i < n; i++) {
		adj_list [i]	= ap;
		tmp_ptr [i]	= ap;
		ap += count [i];
	}
	adj_list [i] = ap;		/* set overall end of list. */

	/* Deposit neighbors into properly allocated lists. */
	ep = &edges [first_edge];
	for (i = first_edge; i < nedges; ep++, i++) {
		ap = tmp_ptr [ep -> p1]++;
		ap -> edge	= i;
		ap -> pnum	= ep -> p2;

		ap = tmp_ptr [ep -> p2]++;
		ap -> edge	= i;
		ap -> pnum	= ep -> p1;
	}

	free ((char *) count);
	free ((char *) tmp_ptr);

	return (adj_list);
}

/*
 * This routine recursively walks through the given Minimum Spanning Tree
 * starting at the given node, and determines the longest edge along the
 * path from the original given root node, to every other node in the tree.
 *
 * Note: this routine assumes that edges are listed in increasing order,
 * just as Kruskal's MST algorithm emits them.
 */

	static
	void
walk (

int			node,		/* IN - current node in tree */
int			longest,	/* IN - longest edge # in cur path */
struct bsd *		bsdp,		/* IN - BSD data structure */
int *			rowp		/* OUT - array of longest edge #'s */
)
{
struct mstadj *		ap;
struct mstadj *		endp;

	/* We now know the longest edge along the path from the root */
	/* to this current node!  Remember it in the matrix! */
	rowp [node] = longest;

	bsdp -> mark [node] = TRUE;
	ap	= bsdp -> adj_list [node];
	endp	= bsdp -> adj_list [node + 1];
	while (ap < endp) {
		if (NOT (bsdp -> mark [ap -> pnum])) {
			walk (ap -> pnum,
			     (longest >= ap -> edge) ? longest : ap -> edge,
			      bsdp,
			      rowp);
		}
		++ap;
	}
	bsdp -> mark [node] = FALSE;
}
