/***********************************************************************

	File:	emst.c
	Rev:	a-1
	Date:	11/10/98

	Copyright (c) 1993, 1999 by Martin Zachariasen & David M. Warme

************************************************************************

	Routines to compute Euclidean Minimum Spanning Trees.

************************************************************************

	Modification Log:

	a-1:	11/10/98	martinz
		: Created.

************************************************************************/

#include "dsuf.h"
#include "steiner.h"

#define ANSI_DECLARATORS
#define REAL coord_t
#include "triangle.h"

/*
 * Global Routines
 */

int			euclidean_mst (struct pset *, struct edge *);
dist_t			euclidean_mst_length (struct pset *);


/*
 * External References
 */

	/* none */


/*
 * Local Routines
 */

static int		build_edges (struct pset *,
				     struct edge **,
				     int);

/*
 * This routine computes the total length of the Euclidean Minimum
 * Spanning Tree of the given set of points.
 */

	dist_t
euclidean_mst_length (

struct pset *		pts	/* IN - point set to find MST length of. */
)
{
int		i;
int		nedges;
dist_t		total;
struct edge *	ep;
struct edge *	edges;

	edges = NEWA (pts -> n - 1, struct edge);

	nedges = euclidean_mst (pts, &edges [0]);

	if (nedges NE pts -> n - 1) {
		fatal ("euclidean_mst_length: Bug 1.");
	}

	total = 0;
	ep = &edges [0];
	for (i = 0; i < nedges; i++) {
		total += ep -> len;
		++ep;
	}

	free ((char *) edges);

	return (total);
}

/*
 * This routine computes an Euclidean Minimum Spanning Tree for the
 * given point set.  The output is a list of edges.
 *
 * The "Triangle" package by Jonathan Richard Shewchuk is used
 * to provide the Delaunay triangulation for the set of points.
 */

	int
euclidean_mst (

struct pset *		pts,		/* IN - point set. */
struct edge *		edges		/* OUT - edge list. */
)
{
int		nedges;
int		mst_edge_count;
struct edge *	edge_array;

	nedges = build_edges (pts, &edge_array, 0);

	mst_edge_count = mst_edge_list (pts -> n,
					nedges,
					&edge_array [0],
					edges);

	free ((char *) edge_array);

	return (mst_edge_count);
}

/*
 * This routine builds an edge-list containing all edges in
 * the Delaunay triangulation of the set of points.
 */

	static
	int
build_edges (

struct pset *		pts,		/* IN - set of points */
struct edge **		edges_out,	/* OUT - edge list */
int			at_least	/* IN - allocate at least this many */
					/*	edges */
)
{
int			i;
int			j;
int             	k;
int             	n;
int	 		nedges;
int			nalloc;
struct edge *		edges;
struct point *		p1;
struct point *		p2;
struct triangulateio	in, out, vorout;

	n = pts -> n;

	/* Set up data structure to call triangle. */
	in.numberofpoints = pts -> n;
	in.pointlist	  = NEWA (2*n, coord_t);
	for (i = 0; i < n; i++) {
		in.pointlist [2*i  ] = pts -> a [i].x;
		in.pointlist [2*i+1] = pts -> a [i].y;
	}

	in.numberofpointattributes	= 0;
	in.pointattributelist		= 0;
	in.pointmarkerlist		= 0;
	in.numberofsegments		= 0;
	in.numberofholes		= 0;
	in.numberofregions		= 0;
	in.regionlist			= 0;
	out.pointlist			= 0;
	out.edgelist			= 0;
	out.trianglelist		= 0;
	out.neighborlist		= 0;

	triangulate ("zeNBQ", &in, &out, &vorout);

	free (in.pointlist);
	free (out.pointlist);
	free (out.trianglelist);
	free (out.neighborlist);

	nedges = out.numberofedges;

	nalloc = nedges;
	if (nalloc < at_least) {
		nalloc = at_least;
	}
	edges = NEWA (nalloc, struct edge);
	*edges_out = edges;

	for (k = 0; k < nedges; k++) {
		i  = out.edgelist [2*k  ];
		j  = out.edgelist [2*k+1];
		p1 = &(pts -> a [i]);
		p2 = &(pts -> a [j]);

		edges -> len	= EDIST (p1, p2);
		edges -> p1	= ((pnum_t) i);
		edges -> p2	= ((pnum_t) j);
		++edges;
	}

	return (nedges);
}

/*
 * This routine sorts the given edge list in INCREASING order by edge length.
 */

	static
	void
sort_edge_list (

struct edge *		a,	/* IN/OUT - array of edges to be sorted. */
int			n	/* IN - number of elements in array. */
)
{
int		h;
struct edge	tmp;
dist_t		key;
struct edge *	p1;
struct edge *	p2;
struct edge *	p3;
struct edge *	p4;
struct edge *	endp;

	endp = &a [n];

	for (h = 1; h <= n; h = 3*h+1) {
	}

	do {
		h = h / 3;
		p4 = &a [h];
		p1 = p4;
		while (p1 < endp) {
			tmp = *p1;
			key = tmp.len;
			p2 = p1;
			while (TRUE) {
				p3 = (p2 - h);
				if (p3 -> len <= key) break;
				*p2 = *p3;
				p2 = p3;
				if (p2 < p4) break;
			}
			*p2 = tmp;
			++p1;
		}
	} while (h > 1);
}

/*
 * This routine computes the MST of a given list of edges.
 */

	int
mst_edge_list (

int			n,		/* IN - number of vertices */
int			nedges,		/* IN - number of edges */
struct edge *		edge_list,	/* IN - list of edges */
struct edge *		edges		/* OUT - MST edge list */
)
{
int		i;
int		mst_edge_count;
int		components;
int		max_vert;
struct edge *	ep;
struct edge *	ep_endp;
int		root1;
int		root2;
struct dsuf	sets;

	sort_edge_list (edge_list, nedges);

	/* Don't assume that the vertex numbers are well-behaved,	*/
	/* except that they must be non-negative.  We do a quick scan	*/
	/* to determine the largest vertex number and then allocate	*/
	/* a union-find data structure large enough to handle it.  Note	*/
	/* that we then use this union-find data structure in a		*/
	/* completely sparse way -- we only ever access set items for	*/
	/* vertices that are named by an edge.				*/

	max_vert = 1;		/* avoid zero-size union-find... */
	ep = edge_list;
	for (i = 0; i < nedges; i++, ep++) {
		if (ep -> p1 > max_vert) {
			max_vert = ep -> p1;
		}
		if (ep -> p2 > max_vert) {
			max_vert = ep -> p2;
		}
	}

	dsuf_create (&sets, max_vert + 1);

	/* Note that it is not a problem to "makeset" a vertex more	*/
	/* than once...							*/
	ep = edge_list;
	for (i = 0; i < nedges; i++, ep++) {
		dsuf_makeset (&sets, ep -> p1);
		dsuf_makeset (&sets, ep -> p2);
	}

	components = n;
	mst_edge_count = 0;
	ep = edge_list;
	ep_endp = (ep + nedges);

	while (components > 1) {
		if (ep >= ep_endp) {
			/* Ran out of edges before MST complete! */
			fatal ("mst_edge_list: Bug 1.");
		}
		root1 = dsuf_find (&sets, ep -> p1);
		root2 = dsuf_find (&sets, ep -> p2);
		if (root1 NE root2) {
			dsuf_unite (&sets, root1, root2);
			*edges = *ep;
			++edges;
			++mst_edge_count;
			--components;
		}
		++ep;
	}

	dsuf_destroy (&sets);

	return (mst_edge_count);
}
