/***********************************************************************

	File:	genps.c
	Rev:	a-1
	Date:	04/18/93

	Copyright (c) 1993, 1999 by David M. Warme

************************************************************************

	Routines for outputting stuff.

************************************************************************

	Modification Log:

	a-1:	04/18/93	warme
		: Created.  Moved most output stuff into this file
		:  from other places.

************************************************************************/

#include "genps.h"
#include "steiner.h"


/*
 * Global Routines
 */

void		begin_plot (enum plot_size);
void		define_Plot_Terminals (struct pset *, double);
void		draw_segment (struct point *, struct point *);
void		end_plot (char *);
void		overlay_plot_subset (char *,
				     bitmap_t *,
				     struct cinfo *,
				     enum plot_size);
void		plot_full_sets (bitmap_t *, struct cinfo *, enum plot_size);
void		plot_full_sets_grouped (bitmap_t *,
					struct cinfo *,
					enum plot_size);
void		plot_lp_solution (char *,
				  double *,
				  struct cinfo *,
				  enum plot_size);


/*
 * External References
 */

	/* none */


/*
 * Local Equates
 */

#define	SMALL_PLOTS_PER_PAGE	12


/*
 * Local Routines
 */

static void		announce_new_page (void);
static void		define_coordinate_axes (struct pset *, double);
static void		draw_efst (struct full_set *);
static void		draw_fst (struct full_set *, struct cinfo *);
static void		draw_rfst (struct full_set *);
static void		draw_fractional_fst (struct full_set *,
					     double,
					     struct cinfo *);
static void		fst_comment (struct full_set *);
static double		get_limit (double);
static void		page_break (void);


/*
 * Local Variables
 */

static int		current_page = 0;
static enum plot_size	current_plot_size;
static int		small_plots_on_page = 0;

/*
 * This routine emits Postscript that defines the coordinates of
 * all the terminals.  The "N DefineTerminals" procedure causes space
 * to be allocated for N terminals.  We then emit the terminals, one
 * per line with "X Y DT" procedure calls.  The "DT" procedure simply
 * stuffs the given X,Y coordinates into the next slot in the TermX and
 * TermY arrays.
 *
 * Once the terminals are defined, the X,Y coordinates of a terminal (e.g.,
 * terminal 57) can be pushed onto the Postscript stack using "57 T".
 */

	void
define_Plot_Terminals (

struct pset *		pts,		/* IN - terminals to plot */
double			scale_value	/* IN - coordinate scale value */
)
{
int			i;
int			n;
struct point *		p1;
char			buf1 [32];
char			buf2 [32];

	if (pts EQ NULL) return;

	n = pts -> n;

	printf ("\n%%%%BeginSetup\n");

	define_coordinate_axes (pts, scale_value);

	printf ("\n%d DefineTerminals\n", n);

	p1 = &(pts -> a [0]);
	for (i = 0; i < n; i++, p1++) {
		coord_to_string (buf1, p1 -> x);
		coord_to_string (buf2, p1 -> y);
		(void) printf ("\t%s\t%s\tDT\n", buf1, buf2);
	}
	printf ("\n%%%%EndSetup\n\n");
}

/*
 * This routine determines appropriate ranges for the X and Y coordinate
 * axes, and emits the corresponding Postscript definitions for the
 * MinX, MaxX, MinY and MaxY variables.
 */

	static
	void
define_coordinate_axes (

struct pset *		pts,		/* IN - terminals to plot */
double			scale		/* IN - coordinate scale value */
)
{
int			i;
int			n;
struct point *		p1;
double			x, y;
double			minxcoord, maxxcoord;
double			minycoord, maxycoord;
double			axmin, axmax;
double			aymin, aymax;

	n = pts -> n;

	if (n < 1) {
		printf ("\n0 1 0 1 SetAxes\n");
		return;
	}

	p1 = &(pts -> a [0]);

	minxcoord = maxxcoord = p1 -> x / scale;
	minycoord = maxycoord = p1 -> y / scale;
	++p1;

	for (i = 1; i < n; i++, p1++) {
		x = p1 -> x / scale;
		y = p1 -> y / scale;
		if (x < minxcoord) {
			minxcoord = x;
		}
		else if (x > maxxcoord) {
			maxxcoord = x;
		}
		if (y < minycoord) {
			minycoord = y;
		}
		else if (y > maxycoord) {
			maxycoord = y;
		}
	}

	/* Determine the gross magnitudes */
	axmax = fabs (maxxcoord);
	axmin = fabs (minxcoord);
	if (axmax < axmin) {
		axmax = axmin;
	}
	axmax = get_limit (axmax);

	aymax = fabs (maxycoord);
	aymin = fabs (minycoord);
	if (aymax < aymin) {
		aymax = aymin;
	}
	aymax = get_limit (aymax);

	/* The following code assumes the X coordinates are not */
	/* closely packed around a point distant from the origin... */

	if (maxxcoord < 0.0) {
		axmin = - axmax;
		axmax = 0.0;
	}
	else if (minxcoord >= 0.0) {
		axmin = 0.0;
	}
	else {
		axmin = -axmax;
	}
	if (maxycoord < 0.0) {
		aymin = - aymax;
		aymax = 0.0;
	}
	else if (minycoord >= 0.0) {
		aymin = 0.0;
	}
	else {
		aymin = -aymax;
	}

	/* Good enough for now... */
	printf ("\n%g %g %g %g SetAxes\n", axmin, axmax, aymin, aymax);
}

/*
 * This routine returns a reasonable scale limit for the given quantity.
 * These are always 1, 2 or 5 times a power of 10.
 */

	static
	double
get_limit (

double		value		/* IN - value to get scale limit for */
)
{
int		i;
double		limit;

	if (value >= 1.0) {
		limit = 1.0;
		for (i = 0; i <= 20; i++) {
			if (limit >= value) return (limit);
			if (2.0 * limit >= value) return (2.0 * limit);
			if (5.0 * limit >= value) return (5.0 * limit);
			limit *= 10.0;
		}
		return (value);
	}

	limit = 1.0;
	for (i = 0; i <= 20; i++) {
		if (0.5 < value * limit) return (1.0 / limit);
		limit *= 10.0;
		if (2.0 < value * limit) return (5.0 / limit);
		if (1.0 < value * limit) return (2.0 / limit);
	}
	return (value);
}

/*
 * This routine generates some Postscript commands to plot each full-set
 * in the given mask.
 */

	void
plot_full_sets (

bitmap_t *		fset_mask,	/* IN - subset of full-sets to plot */
struct cinfo *		cip,		/* IN - compatibility info */
enum plot_size		plot_size	/* IN - size of plot to produce */
)
{
int			i;
int			n;
struct full_set *	fsp;
int *			vp1;
int *			vp2;
char			title [80];
char			buf1 [32];

	n = cip -> num_edges;

	if ((cip -> pts NE NULL) AND (cip -> full_trees NE NULL)) {
		/* Draw the FSTs with postscript. */
		for (i = 0; i < n; i++) {
			if (NOT BITON (fset_mask, i)) continue;
			fsp = cip -> full_trees [i];
			begin_plot (plot_size);
			(void) printf ("\tPlot_Terminals\n");
			fst_comment (fsp);
			draw_fst (fsp, cip);
			dist_to_string (buf1, cip -> cost [i]);
			(void) sprintf (title, "FST %lu,  Length = %s",
					(int32u) i, buf1);
			end_plot (title);
		}
		page_break ();
	}
	else {
		/* Just print the hyperedges */
		for (i = 0; i < n; i++) {
			if (NOT BITON (fset_mask, i)) continue;
			printf ("Edge %d: ", i);
			vp1 = cip -> edge [i];
			vp2 = cip -> edge [i + 1];
			while (vp1 < vp2) {
				printf ("%d ", *vp1++);
			}
			dist_to_string (buf1, cip -> cost [i]);
			printf ("%s\n", buf1);
		}
	}
}

/*
 * This routine generates some Postscript commands to plot each full-set
 * in the given mask.  Note that we "group" as many mutually-disjoint
 * full sets as possible on each plot so as to minimize the amount of
 * paper we spew...
 */

	void
plot_full_sets_grouped (

bitmap_t *		fset_mask,	/* IN - subset of full-sets to plot */
struct cinfo *		cip,		/* IN - compatibility info */
enum plot_size		plot_size	/* IN - size of plot to produce */
)
{
int			i;
int			j;
int			n;
int			nleft;
int			kmasks;
int			nmasks;
int			nplot;
bitmap_t *		left;
bitmap_t *		tmask;
int *			plist;
struct full_set *	fsp;
struct pset *		pts;
char *			cp1;
char *			cp2;
char			fsnums [110];
char			fsnum [8];
char			title [120];

	n = cip -> num_edges;
	nmasks = cip -> num_edge_masks;
	kmasks = cip -> num_vert_masks;

	left = NEWA (nmasks, bitmap_t);
	tmask = NEWA (kmasks, bitmap_t);
	plist = NEWA (n, int);

	/* Make a local copy of all full sets left to plot. */
	for (i = 0; i < nmasks; i++) {
		left [i] = fset_mask [i];
	}
	nleft = n;

	while (nleft > 0) {
		begin_plot (plot_size);

		for (i = 0; i < kmasks; i++) {
			tmask [i] = 0;
		}

		nplot = 0;
		cp1 = &fsnums [sizeof (fsnums)];
		*--cp1 = '\0';

		for (i = n - 1; i >= 0; i--) {
			if (NOT BITON (left, i)) continue;

			/* Skip full set "i" if not disjoint	*/
			/* with all others in this plot...	*/
			fsp = cip -> full_trees [i];
			pts = fsp -> terminals;
			for (j = 0; j < pts -> n; j++) {
				if (BITON (tmask, pts -> a [j].pnum)) break;
			}
			if (j < pts -> n) continue;

			(void) sprintf (fsnum, "%lu", (int32u) i);
			for (cp2 = fsnum; *cp2 NE '\0'; cp2++) {
			}

			/* Stop if label does not fit! */
			if ((cp2 - fsnum) + 2 > (cp1 - fsnums)) break;

			while (cp2 > fsnum) {
				*--cp1 = *--cp2;
			}
			*--cp1 = ' ';
			*--cp1 = ',';

			plist [nplot++] = i;
			CLRBIT (left, i);
			--nleft;

			fsp = cip -> full_trees [i];
			fst_comment (fsp);
			draw_fst (fsp, cip);

			pts = fsp -> terminals;
			for (j = 0; j < pts -> n; j++) {
				SETBIT (tmask, pts -> a [j].pnum);
			}
		}

		(void) printf ("\tPlot_Terminals\n");
		(void) sprintf (title,
				"FST%s %s.",
				(nplot > 1) ? "s" : "",
				cp1 + 2);
		end_plot (title);
	}

	page_break ();

	free ((char *) plist);
	free ((char *) tmask);
	free ((char *) left);
}

/*
 * This routine generates some Postscript commands to plot a SUBSET of
 * the given full-sets in overlaid fashion.
 */

	void
overlay_plot_subset (

char *			title,		/* IN - title to display with. */
bitmap_t *		fset_mask,	/* IN - subset of full-sets to plot */
struct cinfo *		cip,		/* IN - compatibility info */
enum plot_size		plot_size	/* IN - size of plot to produce */
)
{
int			i;
int			n;
int *			vp1;
int *			vp2;
struct full_set *	fsp;
char			buf1 [32];

	n = cip -> num_edges;

	if ((cip -> pts NE NULL) AND (cip -> full_trees NE NULL)) {
		/* Draw the FSTs with postscript. */
		begin_plot (plot_size);
		(void) printf ("\tPlot_Terminals\n");

		for (i = 0; i < n; i++) {
			if (NOT BITON (fset_mask, i)) continue;
			fsp = cip -> full_trees [i];
			fst_comment (fsp);
			draw_fst (fsp, cip);
		}

		end_plot (title);
	}
	else {
		/* Just print the hyperedges */
		for (i = 0; i < n; i++) {
			if (NOT BITON (fset_mask, i)) continue;
			printf ("Edge %d: ", i);
			vp1 = cip -> edge [i];
			vp2 = cip -> edge [i + 1];
			while (vp1 < vp2) {
				printf ("%d ", *vp1++);
			}
			dist_to_string (buf1, cip -> cost [i]);
			printf ("%s\n", buf1);
		}
	}
}

/*
 * This routine draws a single full Steiner tree.
 */

	static
	void
draw_fst (

struct full_set *	fsp,		/* IN - FST to draw */
struct cinfo *		cip		/* IN - compatibility info */
)
{
	switch (cip -> metric) {
	case RECTILINEAR:
		draw_rfst (fsp);
		break;

	case EUCLIDEAN:
		draw_efst (fsp);
		break;

	default:
		fatal ("draw_fst: Bug 1.");
		break;
	}
}

/*
 * This routine draws a single rectilinear FST.
 *
 * NOTE:  We ASSUME that the given FST graph represents a left-most Hwang
 * topology!  (Beyond this, we don't really care about the ordering of
 * terminals, edges or Steiner points.)  If we therefore draw all
 * "diagonal" (i.e., neither vertical nor horizontal) edges in left-most
 * fashion (vertical segment to left of the horizontal segment), then the
 * FST will be correctly drawn as a left-most Hwang topology.
 */

	static
	void
draw_rfst (

struct full_set *	fsp		/* IN - full set to draw */
)
{
int			i;
int			nt;
struct point *		p1;
struct point *		p2;
struct pset *		terms;
struct pset *		steins;
struct edge *		ep;
char			cbufx [32];
char			cbufy [32];
char			buf1 [80];
char			buf2 [80];

	terms = fsp -> terminals;
	steins = fsp -> steiners;

	nt = terms -> n;

	ep = fsp -> edges;
	for (i = 0; i < fsp -> nedges; i++, ep++) {
		if (ep -> p1 < nt) {
			p1 = &(terms -> a [ep -> p1]);
			(void) sprintf (buf1, "%d T", p1 -> pnum);
		}
		else {
			p1 = &(steins -> a [ep -> p1 - nt]);
			coord_to_string (cbufx, p1 -> x);
			coord_to_string (cbufy, p1 -> y);
			(void) sprintf (buf1, "%s\t%s", cbufx, cbufy);
		}

		if (ep -> p2 < nt) {
			p2 = &(terms -> a [ep -> p2]);
			(void) sprintf (buf2, "%d T", p2 -> pnum);
		}
		else {
			p2 = &(steins -> a [ep -> p2 - nt]);
			coord_to_string (cbufx, p2 -> x);
			coord_to_string (cbufy, p2 -> y);
			(void) sprintf (buf2, "%s\t%s", cbufx, cbufy);
		}

		if ((p1 -> x EQ p2 -> x) OR (p1 -> y EQ p2 -> y)) {
			(void) printf ("\t%s\t%s\tS\n",
				       buf1,
				       buf2);
		}
		else if (p1 -> x < p2 -> x) {
			(void) printf ("\t%s\t%s\tC\n",
				       buf1,
				       buf2);
		}
		else {
			(void) printf ("\t%s\t%s\tC\n",
				       buf2,
				       buf1);
		}
	}
}

/*
 * This routine draws a single Euclidean FST.
 */

	static
	void
draw_efst (

struct full_set *	fsp		/* IN - FST to draw */
)
{
int			i;
int			j;
int			nt;
struct point *		p1;
struct pset *		terms;
struct pset *		steins;
struct edge *		ep;
char			buf1 [32];
char			buf2 [32];

	terms = fsp -> terminals;
	steins = fsp -> steiners;

	nt = terms -> n;

	ep = fsp -> edges;
	for (i = 0; i < fsp -> nedges; i++, ep++) {
		j = ep -> p1;
		if (j < nt) {
			p1 = &(terms -> a [j]);
			(void) printf ("\t%d T", p1 -> pnum);
		}
		else {
			p1 = &(steins -> a [j - nt]);
			coord_to_string (buf1, p1 -> x);
			coord_to_string (buf2, p1 -> y);
			(void) printf ("\t%s\t%s", buf1, buf2);
		}
		j = ep -> p2;
		if (j < nt) {
			p1 = &(terms -> a [j]);
			(void) printf ("\t%d T\tS\n", p1 -> pnum);
		}
		else {
			p1 = &(steins -> a [j - nt]);
			coord_to_string (buf1, p1 -> x);
			coord_to_string (buf2, p1 -> y);
			(void) printf ("\t%s\t%s\tS\n", buf1, buf2);
		}
	}
}

/*
 * This routine plots an LP solution.  This is a set of full sets in which
 * full set i has weight Wi, where 0 <= Wi <= 1.  The full sets with
 * weight of 1 are drawn normally.  The fractional ones are drawn as
 * gray-scale "stars" emanating from the center of mass of the terminals.
 */

	void
plot_lp_solution (

char *		title,		/* IN - title for plot. */
double *	weights,	/* IN - weight of each full set. */
struct cinfo *	cip,		/* IN - compatibility info. */
enum plot_size	plot_size	/* IN - size of plot to produce. */
)
{
int			i;
int			n;
struct full_set *	fsp;
int *			vp1;
int *			vp2;

	n = cip -> num_edges;

	if ((cip -> pts NE NULL) AND (cip -> full_trees NE NULL)) {
		/* Draw the FSTs with postscript. */
		begin_plot (plot_size);

		for (i = 0; i < n; i++) {
			if (weights [i] <= 0.000001) continue;
			fsp = cip -> full_trees [i];
			draw_fractional_fst (fsp, weights [i], cip);
		}

		(void) printf ("\tPlot_Terminals\n");

		end_plot (title);
	}
	else {
		/* Just output the weighted hyperedges. */
		printf ("\n");
		for (i = 0; i < n; i++) {
			if (weights [i] <= 0.000001) continue;
			printf ("x^d = %g\t", i, weights [i]);
			vp1 = cip -> edge [i];
			vp2 = cip -> edge [i + 1];
			while (vp1 < vp2) {
				printf (" %d", *vp1++);
			}
			printf ("\n");
		}
	}
}

/*
 * This routine draws a single fractional-weight full set.  We draw these
 * as a "star" using the center-of-mass of the terminals as the hub.  The
 * weight is used to determine the darkness of the lines.
 */

	static
	void
draw_fractional_fst (

struct full_set *	fsp,		/* IN - full set to draw */
double			weight,		/* IN - weight for full set */
struct cinfo *		cip		/* IN - compatibility info */
)
{
int			i;
int			n;
coord_t			cx;
coord_t			cy;
struct pset *		terms;
struct point *		p1;
char			buf1 [32];
char			buf2 [32];

	terms = fsp -> terminals;

	if (weight + 0.000001 >= 1.0) {
		/* Draw integral full sets "normally"... */
		fst_comment (fsp);
		(void) printf ("\t0 setgray\n");
		draw_fst (fsp, cip);
		return;
	}

	fst_comment (fsp);

	n = terms -> n;

	/* Determine the coordinates of the "hub". */
	p1 = &(terms -> a [0]);
	cx = 0.0;
	cy = 0.0;
	for (i = 0; i < n; p1++, i++) {
		cx += p1 -> x;
		cy += p1 -> y;
	}
	cx /= n;
	cy /= n;
	coord_to_string (buf1, cx);
	coord_to_string (buf2, cy);

	(void) printf ("\t%f setgray\n", 1.0 - weight);

	p1 = &(terms -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		(void) printf ("\t%d T\t%s\t%s\tS\n",
			       p1 -> pnum, buf1, buf2);
	}
}

/*
 * This routine emits a Postscript comment describing the FST.
 */

	static
	void
fst_comment (

struct full_set *	fsp		/* IN - full set to describe */
)
{
int			i;
int			n;
struct pset *		terms;
struct point *		p1;

	terms = fsp -> terminals;

	(void) printf (" %% fs%d:", fsp -> tree_num);
	n = terms -> n;
	p1 = &(terms -> a [0]);
	for (i = 0; i < n; p1++, i++) {
		(void) printf (" %lu", (int32u) (p1 -> pnum));
	}
	(void) printf ("\n");
}

/*
 * This routine draws a line segment between two points.
 */

	void
draw_segment (

struct point *		p1,	/* IN - first point. */
struct point *		p2	/* IN - second point. */
)
{
char		buf1 [32];
char		buf2 [32];
char		buf3 [32];
char		buf4 [32];

	coord_to_string (buf1, p1 -> x);
	coord_to_string (buf2, p1 -> y);
	coord_to_string (buf3, p2 -> x);
	coord_to_string (buf4, p2 -> y);

	(void) printf ("	%s	%s	%s	%s	S\n",
		       buf1, buf2, buf3, buf4);
}

/*
 * This routine emits appropriate Postscript code to begin a plot of
 * the given size.  If this is the first plot on a page, then we also
 * emit DSC comments for the page number.  This helps out ghostview and
 * other Postscript previewers.
 */

	void
begin_plot (

enum plot_size		size		/* IN - size of plot to begin */
)
{
	current_plot_size = size;
	switch (size) {
	case BIG_PLOT:
		page_break ();
		announce_new_page ();
		(void) printf ("BeginPlot\n");
		break;

	case SMALL_PLOT:
		if (small_plots_on_page EQ 0) {
			announce_new_page ();
		}
		(void) printf ("BeginSmallPlot\n");
		break;

	default:
		fatal ("begin_plot: Bug 1.");
		break;
	}
}

/*
 * This routine emits appropriate Postscript code to end the plot that
 * is currently in progress.  We also track the number of finished
 * plots per page here.
 */

	void
end_plot (

char *			title		/* IN - title for plot */
)
{
	(void) printf ("  (%s)\n", title);
	switch (current_plot_size) {
	case BIG_PLOT:
		(void) printf ("EndPlot\n\n");
		break;

	case SMALL_PLOT:
		(void) printf ("EndSmallPlot2\n\n");
		++small_plots_on_page;
		if (small_plots_on_page >= SMALL_PLOTS_PER_PAGE) {
			small_plots_on_page = 0;
		}
		break;

	default:
		fatal ("end_plot: Bug 1.");
		break;
	}
}

/*
 * This routine puts a %%Page: comment into the output to mark a page
 * boundary.  This is for the benefit of Ghostview and other Postscript
 * previewers.
 */

	static
	void
announce_new_page (void)

{
	++current_page;
	(void) printf ("%%%%Page: %lu %lu\n",
		       (int32u) current_page,
		       (int32u) current_page);
}

/*
 * This routine introduces a page break.  This is needed when doing
 * small plots, and the page has not yet filled up.
 */

	static
	void
page_break (void)

{
	if (small_plots_on_page NE 0) {
		(void) printf ("FlushSmallPlot\n");
		small_plots_on_page = 0;
	}
}
